import os

SUBJECT_FMT = '%s%03d'
"""
The XNAT subject name format with argument collection and subject number.
"""

SESSION_FMT = 'Session%02d'
"""The XNAT session name format with argument session number."""

CONF_DIR = os.path.abspath(os.path.join(os.path.dirname(__file__), '..', 'conf'))
"""The common configuration directory."""

SCAN_TS_RSC = 'scan_ts'
"""The XNAT scan time series resource name."""

MASK_RSC = 'mask'
"""The XNAT mask resource name."""



"""
This module reorders the OHSU AIRC ``bolero_mask_conv``
result to conform with the time series x and y order.
"""
import os
from os import path
from glob import glob
import traits.api as traits
from nipype.interfaces.base import (TraitedSpec, CommandLine,
                                    CommandLineInputSpec)
from nipype.interfaces.traits_extension import Undefined
from qiutil.file import splitexts


class ReorderBoleroMaskInputSpec(CommandLineInputSpec):
    in_file = traits.Str(desc='mask file', mandatory=True,
                         position=1, argstr='\"%s\"')
    out_file = traits.Str(desc='Output file name', argstr='-o %s')


class ReorderBoleroMaskOutputSpec(TraitedSpec):
    out_file = traits.File(desc='Reordered mask file', exists=True)


class ReorderBoleroMask(CommandLine):
    """
    Interface to the mask reordering utility.
    """

    _cmd = 'bolero_mask_reorder'
    input_spec = ReorderBoleroMaskInputSpec
    output_spec = ReorderBoleroMaskOutputSpec

    def __init__(self, **inputs):
        super(ReorderBoleroMask, self).__init__(**inputs)

    def run(self, **inputs):
        if not inputs.get('out_file'):
            in_file = inputs.get('in_file')
            # in_file is mandatory, but let that be checked by the
            # base class and the appropriate error thrown there.
            if in_file:
                inputs['out_file'] = self._default_output_file_name(in_file)
        super(ReorderBoleroMask, self).run(**inputs)

    def _list_outputs(self):
        outputs = self._outputs().get()
        out_file = self.inputs.out_file
        # Expand the output path, if necessary.
        outputs['out_file'] = os.path.abspath(out_file)

        return outputs

    def _default_output_file_name(self, in_file):
        """
        The default output file name appends ``_reordered``
        to the input file base name.
        """
        _, in_file_name = os.path.split(in_file)
        base, ext = splitexts(in_file_name)
        return base + '_reordered' + ext



# Absolute import (the default in a future Python release) resolves
# the logging import as the Python standard logging module rather
# than this module of the same name.
from __future__ import absolute_import
import os
import sys
from datetime import datetime
import logging
import qiutil

NIPYPE_LOG_DIR_ENV_VAR = 'NIPYPE_LOG_DIR'
"""The environment variable used by Nipype to set the log directory."""


def configure(**opts):
    """
    Configures the logger as follows:

    - If there is a *log* option,
      then the logger is a conventional ``qiutil.logging`` logger
      which writes to the given log file.

    - Otherwise, the logger delegates to a mock logger that
      writes to stdout.

    :param opts: the ``qiutil.command.configure_log`` options
    :return: the logger factory
    """
    # The log file option.
    log_file_opt = opts.get('log')
    # Set the Nipype log directory environment variable before importing
    # any nipype module. The code below works around the following Nipype
    # bug:
    # * Nipype requires a log directory. If the Nipype log directory is
    #   set to /dev/null, then Nipype raises an error. The work-around
    #   is to set the NIPYPE_LOG_DIR environment variable to a new temp
    #   directory.
    log_dir = None
    if log_file_opt:
        # Configure the qiutil logger for the qi* modules.
        qiutil.command.configure_log('qipipe', 'qixnat', 'qidicom',
                                     'qiutil', **opts)
        log_file = os.path.abspath(log_file_opt)
        if log_file == '/dev/null':
            # Work around the Nipype bug described above.
            log_dir = tempfile.mkdtemp(prefix='qipipe_')
        else:
            log_dir = os.path.dirname(log_file)
        # Make the log file parent directory, if necessary.
        if not os.path.exists(log_dir):
            os.makedirs(log_dir)
        factory = qiutil.logging.logger
    else:
        # Print log messages to stdout to work around the Nipype
        # bug described in the logger method apidoc.
        mock_log_opts = {}
        level = opts.get('log_level')
        if level:
            mock_log_opts['level'] = level
        factory = MockLoggerFactory(**mock_log_opts).logger

    # Nipype always needs a log directory to work around the
    # following Nipype bug:
    # * If the Nipype log directory is not set, then Nipype still
    #   logs to the default log file ./log/pypeline.log, but also
    #   logs to stdout, which stomps on the qipipe logging.
    if not log_dir:
        log_dir = '/'.join([os.getcwd(), 'log'])
    # Set the Nipype log directory environment variable.
    os.environ[NIPYPE_LOG_DIR_ENV_VAR] = log_dir

    # Set the global logger factory.
    logger._factory = factory
    # Print a log message.
    log_dest = log_file_opt if log_file_opt else 'stdout'
    factory(__name__).info("Logging qipipe to %s." % log_dest)
    factory(__name__).info("Logging nipype to the %s directory." %
                           log_dir)

    return factory

def logger(name):
    """
    This method overrides ``qiutil.logging.logger`` to work
    around the following Nipype bug:

    * Nipype stomps on any other application's logging.
      The work-around is to mock a "logger" that writes
      to stdout.

    :param name: the caller's context ``__name__``
    :return: the logger facade
    """
    # Make a default logger factory on demand.
    if not logger._factory:
        logger._factory = configure()

    return logger._factory(name)


class MockLoggerFactory(object):
    def __init__(self, **opts):
        self.writer = MockLogWriter(**opts)

    def logger(self, name):
        return MockLogger(self.writer, name)


class MockLogger(object):
    def __init__(self, writer, name):
        self.writer = writer
        self.name = name

    @property
    def level(self):
        return self.writer.level

    def info(self, message):
        self.writer.info(self.name, message)

    def error(self, message):
        self.writer.error(self.name, message)

    def warn(self, message):
        self.writer.warn(self.name, message)

    def debug(self, message):
        self.writer.debug(self.name, message)


class MockLogWriter(object):
    def __init__(self, level=None):
        if not level:
            level = 'INFO'
        self.level_s = level
        self.level = getattr(logging, self.level_s)

    def info(self, name, message):
        if self.level <= logging.INFO:
            self._write(name, message)

    def error(self, name, message):
        if self.level <= logging.ERROR:
            self._write(name, message)

    def warn(self, name, message):
        if self.level <= logging.WARN:
            self._write(name, message)

    def debug(self, name, message):
        if self.level <= logging.DEBUG:
            self._write(name, message)

    def _write(self, name, message):
        dt = datetime.now().strftime("%M/%D/%Y %H:%M:%S")
        print "%s %s %s %s" % (dt, name, self.level_s, message)
        sys.stdout.flush()



"""
OHSU - ROI utility functions.

TODO - move this to ohsu-qipipe.
"""

import os
import re
import glob
import qiutil


class ROIError(Exception):
    pass


class LesionROI(object):
    """
    Aggregate with attributes :attr:`lesion` :attr:`volume`,
    :attr:`slice` and :attr:`location`.
    """
    def __init__(self, lesion, volume_number, slice_sequence_number,
                 location):
        """
        :param lesion: the :attr:`lesion` value
        :param volume_number: the :attr:`volume` value
        :param slice_sequence_number: the :attr:`slice` value
        :param location: the :attr:`location` value
        """
        self.lesion = lesion
        """The lesion number."""

        self.volume = volume_number
        """The one-based volume number."""

        self.slice = slice_sequence_number
        """The one-based slice sequence number."""

        self.location = location
        """The absolute BOLERO ROI .bqf file path."""

    def __repr__(self):
        return (self.__class__.__name__ +
                str(dict(lesion=self.lesion, volume=self.volume,
                         slice=self.slice, location=self.location)))


PARAM_REGEX = re.compile('(?P<key>\w+)\s*\:\s*(?P<value>\w+)')
"""
The regex to parse a parameter file.
"""


def iter_roi(glob_pat, regex, input_dir):
    """
    Iterates over the the OHSU ROI ``.bqf`` mask files in the given
    input directory. This method is a :class:`LesionROI` generator,
    e.g.::

        >>> # Find .bqf files anywhere under /path/to/session/processing.
        >>> next(iter_roi('processing/*', '.*/\.bqf', '/path/to/session'))
        {lesion: 1, slice: 12, path: '/path/to/session/processing/rois/roi.bqf'}

    :param glob_pat: the glob match pattern
    :;param regex: the file name match regular expression
    :param input_dir: the source session directory to search
    :yield: the :class:`LesionROI` objects
    """
    finder = qiutil.file.Finder(glob_pat, regex)
    for match in finder.match(input_dir):
        file_name = match.group(0)
        # If there is no lesion qualifier, then there is only one lesion.
        try:
            lesion_s = match.group('lesion')
        except IndexError:
            lesion_s = None
        lesion = int(lesion_s) if lesion_s else 1

        # Prepend the full input directory to the file match.
        abs_input_dir = os.path.abspath(input_dir)
        file_path = '/'.join([abs_input_dir, file_name])
        # The volume and slice are in the companion parameters file.
        roi_dir, _ = os.path.split(file_path)
        param_file_pat = "%s/*.par" % roi_dir
        param_file_names = glob.glob(param_file_pat)
        if not param_file_names:
            raise ROIError("The ROI slice directory does not have" +
                           " a parameter file: %s" % roi_dir)
        if len(param_file_names) > 1:
            raise ROIError("The ROI slice directory has more than" +
                           "one parameter file: %s" % roi_dir)
        param_file_name = param_file_names[0]
        params = _collect_parameters(param_file_name)

        # If there is no slice number, then complain.
        slice_seq_nbr_s = params.get('CurrentSlice')
        if not slice_seq_nbr_s:
            raise ROIError("The ROI slice could not be determined from" +
                           " the parameter file: %s" % param_file_name)
        slice_seq_nbr = int(slice_seq_nbr_s)

        # If there is no volume number, then complain.
        volume_nbr_s = params.get('CurrentTimePt')
        if not volume_nbr_s:
            raise ROIError("The ROI volume could not be determined from" +
                           " the parameter file: %s" % param_file_name)
        volume_nbr = int(volume_nbr_s)

        yield LesionROI(lesion, volume_nbr, slice_seq_nbr, file_path)

def _collect_parameters(param_file_name):
    """
    Parses the parameters from the ``.par`` file.

    :param param_file_name: the paramter file location
    :return: the parameter {name: value} dictionary
    """
    params = {}
    with open(param_file_name) as f:
        lines = f.readlines()
        for line in lines:
            match = PARAM_REGEX.match(line)
            if match:
                key = match.group('key')
                value = match.group('value')
                params[key] = value
    return params



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft = python sts = 4 ts = 4 sw = 4 et:
"""ANTS Affine Initializer interface
   
   Change directory to provide relative paths for doctests
   >>> import os
   >>> filepath = os.path.dirname( os.path.realpath( __file__ ) )
   >>> datadir = os.path.realpath(os.path.join(filepath,
   ...                                         '../../../test/fixtures/staged/'
   ...                                         'breast/Breast003/Session01/'
   ...                                         'scans/1/resources'))
   >>> os.chdir(datadir)

"""
import os

from nipype.interfaces.ants.base import (ANTSCommand, ANTSCommandInputSpec)
from nipype.interfaces.base import (TraitedSpec, File, traits)
from nipype.interfaces.base import InputMultiPath


class AffineInitializerInputSpec(ANTSCommandInputSpec):
    """
    :Note: The antsAffineInitializer command line options are positional
        rather than keyword, e.g.::
            
            antsAffineInitializer 3 fixed.nii.gz moving.nii.gz initial_xform.mat \
            10 0.3
        
        rather than::
            
            antsAffineInitializer 3 fixed.nii.gz moving.nii.gz initial_xform.mat \
            --search_factor 10 --radian_fraction 0.3
        
        Therefore, each optional input field requires the prior input field,
        e.g. *radian_fraction* requires a *search_factor* input. Reasonable
        starting option values are shown in the :class:`AffineInitializer`
        example.
        
        In particular, note that since an image mask is the last positional
        argument, if the *image_mask* input option is set, then all other
        options must be set as well.
    """
    dimension = traits.Enum(3, 2, argstr='%d', usedefault=False,
                            mandatory=True, position=0,
                            desc='image dimension (2 or 3)')
    fixed_image = File(argstr='%s', exists=True, mandatory=True, position=1,
                       desc='realignment reference image')
    moving_image = File(argstr='%s', exists=True, mandatory=True, position=2,
                        desc='image to be realigned')
    output_affine_transform = File(argstr='%s', mandatory=True, position=3,
                                   desc='the name of the resulting transform')
    search_factor = traits.Int(argstr='%d', position=4,
                               desc='search increments in degrees')
    radian_fraction = traits.Float(argstr='%f', position=5,
                                   requires=['search_factor'],
                                   desc='search fraction about the principal'
                                        ' axis, between 0 and 1')
    use_principal_axes = traits.Bool(argstr='%d', position=6,
                                     requires=['radian_fraction'],
                                     desc='whether the rotation is searched'
                                          ' around an initial principal axis'
                                          ' alignment')
    use_local_search = traits.Int(argstr='%d', position=7,
                                  requires=['use_principal_axes'],
                                  desc='the number of local optimization'
                                       ' iterations is run at each search'
                                       ' point')
    image_mask = File(argstr='%s', exists=True, position=8,
                      requires=['use_local_search'],
                      desc='mask constraining the initializer')


class AffineInitializerOutputSpec(TraitedSpec):
    affine_transform = File(exists=True, desc='output transform file')


class AffineInitializer(ANTSCommand):
    """
    Examples
    --------
    
    >>> initalizer = AffineInitializer()
    >>> initalizer.inputs.dimension = 3
    >>> initalizer.inputs.fixed_image = 'NIFTI/volume003.nii.gz'
    >>> initalizer.inputs.moving_image = 'NIFTI/volume004.nii.gz'
    >>> initalizer.inputs.output_affine_transform = 'initial_xform.mat'
    >>> initalizer.inputs.search_factor = 10
    >>> initalizer.inputs.radian_fraction = 0.3
    >>> initalizer.inputs.use_principal_axes = True
    >>> initalizer.inputs.use_local_search = 20
    >>> initalizer.inputs.image_mask = 'mask/mask.nii.gz'
    >>> initalizer.cmdline
    'antsAffineInitializer 3 NIFTI/volume003.nii.gz NIFTI/volume004.nii.gz initial_xform.mat 10 0.3 1 20 mask/mask.nii.gz'
    """
    _cmd = 'antsAffineInitializer'
    input_spec = AffineInitializerInputSpec
    output_spec = AffineInitializerOutputSpec
    
    
    def _format_arg(self, opt, spec, val):
        if opt == 'radian_fraction':
            # Suppress extraneous trailing zeros.
            # Cf. http://stackoverflow.com/questions/2440692/formatting-floats-in-python-without-superfluous-zeros
            return ('%f' % val).rstrip('0').rstrip('.')
        else:
            return super(AffineInitializer, self)._format_arg(opt, spec, val)
    
    def _list_outputs(self):
        outputs = self._outputs().get()
        out_file = os.path.abspath(self.inputs.output_affine_transform)
        outputs['affine_transform'] = out_file
        return outputs



"""
This module updates the qiprofile database imaging information
from a XNAT scan.
"""
import csv
import tempfile
from qiutil.file import splitexts
from qiutil.collections import concat
from qiutil.ast_config import read_config
from qixnat.helpers import (xnat_path, xnat_name, parse_xnat_date)
from qirest_client.helpers import database
from qirest_client.model.imaging import (
    Session, SessionDetail, Scan, Volume, ScanProtocol, Registration,
    RegistrationProtocol, Modeling, ModelingProtocol
)
from ..staging import image_collection
from ..helpers.constants import (SUBJECT_FMT, SESSION_FMT)
from ..helpers.colors import label_map_basename
from ..pipeline.staging import (SCAN_METADATA_RESOURCE, SCAN_CONF_FILE)
from ..pipeline.registration import (REG_PREFIX, REG_CONF_FILE)
from ..pipeline.modeling import (MODELING_PREFIX, MODELING_CONF_FILE)
from ..pipeline.modeling import INFERRED_R1_0_OUTPUTS as OUTPUTS
from ..pipeline.roi import ROI_RESOURCE
from . import modeling


class ImagingUpdateError(Exception):
    """``qiprofile`` imaging update error."""
    pass


def update(subject, experiment, **opts):
    """
    Updates the imaging content for the qiprofile REST Subject
    from the XNAT experiment.
    
    :param collection: the :attr:`qipipe.staging.image_collection.name``
    :param subject: the target qiprofile Subject to update
    :param experiment: the XNAT experiment object
    :param opts: the :class`Updater` keyword arguments
    """
    Updater(subject, **opts).update(experiment)


class Updater(object):
    def __init__(self, subject, **opts):
        """
        :param subject: the XNAT :attr:`subject`
        :param opts: the following keyword arguments:
        :option bolus_arrival_index: the :attr:`bolus_arrival_index`
        :option roi_centroids: the :attr:`roi_centroids`
        :option roi_average_intensities: the :attr:`roi_average_intensities`
        """
        self.subject = subject
        """The target qiprofile Subject to update."""
        
        self.bolus_arrival_index = opts.get('bolus_arrival_index')
        """
        The scan
        :meth:qipipe.pipeline.qipipeline.bolus_arrival_index_or_zero`.
        """
        
        self.roi_centroids = opts.get('roi_centroids')
        """The ROI centroids list in lesion number order."""
        
        self.roi_average_intensities = opts.get('roi_average_intensities')
        """The ROI average signal intensities list in lesion number order."""
    
    def update(self, experiment):
        """
        Updates the imaging content for the qiprofile REST Subject
        from the XNAT experiment.
        
        :param experiment: the XNAT experiment object
        :raise ImagingUpdateError: if the XNAT experiment does not have
            a visit date
        :raise ImagingUpdateError: if the ``qiprofile`` REST database session
            with the same visit date already exists
        """
        # The XNAT experiment must have a date. The XNAT date is
        # unfortunately a string, which must be converted to a
        # datetime.
        date_s = experiment.attrs.get('date')
        if not date_s:
            raise ImagingUpdateError( "The XNAT experiment %s is missing the"
                                " visit date" % xnat_path(experiment))
        date = parse_xnat_date(date_s)
        # If there is a qiprofile session with the same date,
        # then complain.
        if any( sess.date == date for sess in self.subject.sessions):
            raise ImagingUpdateError(
                "a qiprofile %s %s Subject %d session with visit date %s"
                " already exists" % (self.subject.project,
                                     self.subject.collection,
                                     self.subject.number, date)
            )
        # Make the qiprofile Session object.
        session = self._create_session(experiment)
        # Add the session to the subject encounters in date order.
        self.subject.add_encounter(session)
        # Save the session detail.
        session.detail.save()
        # Save the subject.
        self.subject.save()
    
    def _create_session(self, experiment):
        """
        Makes the qiprofile Session object from the given XNAT experiment.
        
        :param experiment: the XNAT experiment object
        :return: the qiprofile Session object
        """
        # Make the qiprofile scans.
        scans = [self._create_scan(xnat_scan)
                 for xnat_scan in experiment.scans()]
        
        # The modeling resources begin with 'pk_'.
        xnat_mdl_rscs = (rsc for rsc in xnat_scan.resources()
                         if xnat_name(rsc).startswith(MODELING_PREFIX))
        modelings = [self._create_modeling(rsc, scans) for rsc in xnat_mdl_rscs]
        
        # The session detail database object to hold the scans.
        detail = SessionDetail(scans=scans)
        # Save the detail first, since it is not embedded and we need to
        # set the detail reference to make the session.
        detail.save()
        
        # The qiprofile session date is the XNAT experiment date.
        # See the comment in the update method in regards to the
        # type conversion.
        date_s = experiment.attrs.get('date')
        date = parse_xnat_date(date_s)
        
        # Return the new qiprofile Session object.
        return Session(date=date, modelings=modelings, detail=detail)
    
    def _create_scan(self, xnat_scan):
        """
        Makes the qiprofile Session object from the XNAT scan.
        
        :param xnat_scan: the XNAT scan object
        :return: the qiprofile scan object
        """
        # The scan number.
        number = int(xnat_name(xnat_scan))
        # The image collection.
        collection = image_collection.with_name(self.subject.collection)
        # Determine the scan type from the collection and scan number.
        scan_type = collection.scan_types.get(number)
        if not scan_type:
            raise ImagingUpdateError(
                "The %s XNAT scan number %s is not recognized" %
                (self.subject.collection, number)
            )
        # The scan protocol database key.
        pcl_key = dict(technique=scan_type)
        # The corresponding qiprofile ScanProtocol.
        protocol = database.get_or_create(ScanProtocol, pcl_key)
        
        # There must be a time series.
        rsc = xnat_scan.resource('scan_ts')
        if not rsc.exists():
            raise ImagingUpdateError("The XNAT scan %s does not have a time series"
                                     " resource" % xnat_path(xnat_scan))
        time_series_file = next(f for f in rsc.files())
        time_series = TimeSeries(name=xnat_name(time_series_file)
        
        # There must be a bolus arrival.
        if not self.bolus_arrival_index:
            raise ImagingUpdateError("The XNAT scan %s qiprofile update is"
                                    " missing the bolus arrival" %
                                    xnat_path(xnat_scan))
        
        # The ROIs.
        rois = self._create_rois(xnat_scan)
        
        # The XNAT registration resources begin with reg_.
        registrations = [self._create_registration(rsc)
                         for rsc in xnat_scan.resources()
                         if xnat_name(rsc).startswith(REG_PREFIX)]
        
        # Return the new qiprofile Scan object.
        return Scan(number=number, protocol=protocol, time_series=time_series,
                    bolus_arrival_index=self.bolus_arrival_index,
                    rois=rois, registrations=registrations)
    
    def _create_rois(self, xnat_scan, **opts):
        """
        :param xnat_scan: the XNAT scan object
        :param opts: the following keyword arguments:
        :option color_table: the color table file base name
        :option centroids: the ROI centroids list
        :option average_intensities: the ROI average intentsity list
        :return: the qiprofile Regions list for each lesion in
            lesion number order
        """
        # The ROI resource.
        rsc = xnat_scan.resource(ROI_RESOURCE)
        # The [(mask, label map)] list.
        rois = []
        if not rsc.exists():
            return rois
        # The file object label is the file base name.
        fnames = set(rsc.files().get())
        # The mask files do not have _color in the base name.
        # Sort the mask files by the lesion prefix.
        masks = sorted(f for f in fnames if not '_color' in f)
        centroids = opts.get('centroids')
        avg_intensities = opts.get('average_intensities')
        # The color look-up table.
        color_table = opts.get('color_table')
        # Make the ROIs.
        rois = []
        for i, mask in enumerate(masks):
            roi_opts = lobel_map_opts.clone()
            label_map_name = label_map_basename(mask)
            if label_map_name in fnames:
                # A label map must have a color table.
                if not color_table:
                    raise ImagingUpdateError(
                        "The %s label map %s is missing a color table" %
                        (xnat_path(xnat_scan), label_map_name)
                    )
                label_map = LabelMap(filename=label_map_name,
                                     color_table=color_table)
                roi_opts['label_map'] = label_map
            if centroids and i < len(centroids):
                centroid = centroids[i]
                if centroid:
                    roi_opts['centroid'] = centroid
            if avg_intensities and i < len(avg_intensities):
                avg_intensity = avg_intensities[i]
                if avg_intensity:
                    roi_opts['avg_intensity'] = avg_intensity
            roi = Region(mask, **roi_opts)
            rois.append((mask, label_map))
        
        return rois
    
    def _create_registration(self, resource):
        """
        Makes the qiprofile Registration object from the XNAT resource.
        
        :param resource: the XNAT registration resource object
        :return: the qiprofile Registration object
        """
        # Find or create the registration protocol.
        cfg = self._read_configuration(resource, REG_CONF_FILE)
        technique_section = cfg.get('General')
        if not technique_section:
            raise ImagingUpdateError("The XNAT resource %s protocol is"
                                     " missing the General section" %
                                     xnat_path(resource))
        technique = technique_section.get('technique')
        if not technique:
            raise ImagingUpdateError("The XNAT resource %s protocol is"
                                     " missing the technique option" %
                                     xnat_path(resource))
        key = dict(technique=technique, configuration=cfg)
        protocol = database.get_or_create(RegistrationProtocol, key)
        
        # There must be a time series.
        time_series_file = next(f for f in resource.files() if f.label().ends_with('_ts.nii.gz'))
        time_series = TimeSeries(name=xnat_name(time_series_file))
        
        # Return the new qiprofile Registration object.
        return Registration(protocol=protocol, time_series=time_series,
                            resource=xnat_name(resource))
    
    def _create_modeling(self, resource, scans):
        """
        Creates the qiprofile Modeling object from the given XNAT
        resource object.
        
        :param resource: the modeling source XNAT resource object
        :param scans: the REST Scan list for the current session
        :return: the qiprofile Modeling object
        """
        # The XNAT modeling files.
        xnat_files = resource.files()
        
        # The modeling configuration.
        cfg = self._read_configuration(resource, MODELING_CONF_FILE)
        
        # Tease out the modeling source.
        source_topic = cfg.pop('Source', None)
        if not source_topic:
            raise ImagingUpdateError("The XNAT modeling configuration %s is"
                               " missing the Source topic" % xnat_path(cfg_file))
        source_rsc = source_topic.get('resource')
        if not source_rsc:
            raise ImagingUpdateError("The XNAT modeling configuration %s Source"
                               " topic is missing the resource option" %
                               xnat_path(cfg_file))
        xnat_scan = resource.parent()
        if source_rsc.startswith('reg_'):
            reg_lists = (s.registrations for s in scans)
            regs = concat(*reg_lists)
            reg = next((r for r in regs if r.resource == source_rsc), None)
            if not reg:
                raise ImagingUpdateError("The XNAT modeling resource %s source"
                                         " registration resource %s was not"
                                         " found in the session scans" %
                                         (xnat_path(resource), source_rsc))
            source = Modeling.Source(registration=reg.protocol)
        else:
            scan_nbr = xnat_name(xnat_scan)
            scan = next(s for s in scans if s.number == scan_nbr)
            source = Modeling.Source(scan=scan.protocol)
        
        # The qiprofile ModelingProtocol.
        key = dict(technique='Mock', configuration=cfg)
        protocol = database.get_or_create(ModelingProtocol, key)
        
        # The modeling result files.
        xnat_file_labels = {xnat_name(xnat_file) for xnat_file in xnat_files}
        result = {}
        for output in OUTPUTS:
            fname = output + '.nii.gz'
            if fname in xnat_file_labels:
                # TODO - add the param result average and label map to the
                # pipeline and here.
                param_result = Modeling.ParameterResult(filename=fname)
                result[output] = param_result
        
        # Return the new qiprofile Modeling object.
        return Modeling(protocol=protocol, source=source,
                        resource=xnat_name(resource), result=result)
    
    def _read_configuration(self, resource, name):
        """
        :param resource: the XNAT resource configuration file parent
        :param name: the configuration XNAT file name
        :return: the configuration dictionary
        :raise ImagingUpdateError: if the configuration XNAT file was
            not found in the resource
        """
        # The registration configuration.
        cfg_file_finder = (xnat_file for xnat_file in resource.files()
                           if xnat_name(xnat_file) == name)
        xnat_cfg_file = next(cfg_file_finder, None)
        if not xnat_cfg_file:
            raise ImagingUpdateError("The XNAT resource %s does not contain"
                               " the configuration file %s" %
                               (xnat_path(resource), name))
        if xnat_cfg_file:
            cfg_file = xnat_cfg_file.get()
            cfg_dict = dict(read_config(cfg_file))
            return cfg_dict



import re
from .staging_error import StagingError


class ROIPatterns(object):
    """The ROI file name matching patterns."""
    
    def __init__(self, glob, regex):
        self.glob = glob
        """The ROI file glob pattern."""
        
        self.regex = regex
        """The ROI file name match regular expression."""
    
    def __repr__(self):
        return str(dict(glob=self.glob, regex=self.regex))


class ScanPatterns(object):
    """The scan file name matching patterns."""
    
    def __init__(self, dicom, roi=None):
        self.dicom = dicom
        """The DICOM file match *glob* and *regex* patterns."""
        
        self.roi = roi
        """The :class:`ROIPatterns` object."""
    
    def __repr__(self):
        return str(dict(dicom=self.dicom, roi=self.roi))


class Patterns(object):
    """The collection file name and DICOM tag patterns."""
    
    def __init__(self, subject, session, scan, volume):
        """
        :param subject: the subject directory name match regular expression
        :param session: the session directory name match regular expression
        :param scan the {scan number: :class:`ScanPatterns} dictionary
        :option volume: the DICOM tag which identifies a scan volume
        """
        self.subject = subject
        """The subject directory match pattern."""
        
        self.session = session
        """The subject directory match pattern."""
        
        self.scan = scan
        """The {scan number: :class:`ScanPatterns`} dictionary."""
        
        self.volume = volume
        """The DICOM tag which identifies a scan volume."""


def with_name(name):
    """
    :return: the :class:`Collection` whose name is a case-insensitive
        match for the given name, or None if no match is found
    """
    return Collection.instances.get(name.lower())


class Collection(object):
    """The image collection."""
    
    instances = {}
    """The collection {name: object} dictionary."""
    
    def __init__(self, name, **opts):
        """
        :param name: `self.name`
        :param opts: the :class:`Patterns` attributes as well as the
            following keyword options:
        :keyword crop_posterior: the :attr:`crop_posterior` flag
        """
        self.name = name.capitalize()
        """The capitalized collection name."""
        
        self.crop_posterior = opts.pop('crop_posterior', False)
        """
        A flag indicating whether to crop the image posterior in the
        mask, e.g. for a breast tumor (default False).
        """
        
        self.scan_types = opts.pop('scan_types', {})
        """
        The scan {number: type} dictionary.
        """
        
        self.patterns = Patterns(**opts)
        """The file and DICOM meta-data patterns."""
        
        # The scan pattern key is the scan number.
        # There must be a scan type for each number.
        if self.patterns.scan:
            for scan_number in self.patterns.scan:
                if not scan_number in self.scan_types:
                    raise StagingError("The scan number %d is missing a"
                                       " scan type" % scan_number)
        
        # If a collection with this name has not yet been recorded,
        # then add this canonical instance to the collection extent.
        key = name.lower()
        if key not in Collection.instances:
            Collection.instances[key] = self



import os
import csv
from collections import Counter
from qiutil.ast_config import (read_config, ASTConfig)

EXCLUDED_OPTS =  {'plugin_args', 'run_without_submitting'}
"""The config options to exclude in the profile."""


class MetadataError(Exception):
    """Metadata parsing error."""
    pass


def create_profile(configuration, sections, dest, **opts):
    """
    Creates a metadata profile from the given configuration.
    The configuration input is a {section: {option: value}}
    dictionary. The target profile is a Python configuration
    file which includes the given sections. The section content
    is determined by the input configuration and the keyword
    arguments. The keyword item overrides a matching input
    configuration item. The resulting profile is written to a
    new file.
    
    :param configuration: the configuration dictionary
    :param sections: the target profile sections
    :param dest: the target profile file location
    :param opts: additional {section: {option: value}} items
    :return: the target file location
    """
    
    # The target profile to populate.
    profile = ASTConfig()
    # MongoDB does not allow dotted dictionary keys.
    section_map = {section: section.split('.')[-1] for section in sections}
    section_counts = Counter(section_map.itervalues())
    dups = [section for section, count in section_counts.items()
            if count > 1]
    if dups:
        raise MetadataError("The profile configuration has duplicate option"
                            " base names: %s" % dups)
    
    # Add the input parameter sections.
    for cfg_section, prf_section in section_map.iteritems():
        cfg_items = configuration.get(cfg_section)
        if cfg_items:
            # The profile {option, value} dictionary.
            prf_items = {opt: val for opt, val in cfg_items.iteritems()
                         if opt not in EXCLUDED_OPTS}
            if prf_items:
                profile.add_section(prf_section)
                for opt, val in prf_items.iteritems():
                    profile.set(prf_section, opt, val)
    
    # The keyword arguments override the config file.
    for key, items in opts.iteritems():
        # Topics are capitalized.
        section = key.capitalize()
        if not profile.has_section(section):
            profile.add_section(section)
        for opt, val in items.iteritems():
            profile.set(section, opt, val)
    
    # Save the profile.
    dest = os.path.abspath(dest)
    dest_dir = os.path.dirname(dest)
    if not os.path.exists(dest_dir):
        os.makedirs(dest_dir)
    with open(dest, 'w') as f:
        profile.write(f)
    
    return dest



"""Command qipipe command options."""

import qiutil
from qiutil.command import add_options


def configure_log(**opts):
    """
    Configure the logger for the qi* modules.
    """
    qiutil.command.configure_log('qipipe', 'qixnat', 'qidicom', 'qiutil',
                                 **opts)



import os
import re
import glob
import warnings
from setuptools import (setup, find_packages)

VCS_RQMT_PAT = re.compile('^\w+\+\w+:')
"""
The pattern for detecting a VCS requirement spec, e.g.
``git+git://...``.
"""


class InstallError(Exception):
    """qipipe installation error."""
    pass


def version(package):
    """
    :return: the package version as listed in the package `__init.py__`
        `__version__` variable.
    """
    # The version string parser.
    REGEXP = re.compile("""
       __version__   # The version variable
       \s*=\s*       # Assignment
       ['\"]         # Leading quote
       (.+)          # The version string capture group
       ['\"]         # Trailing quote
    """, re.VERBOSE)

    with open(os.path.join(package, '__init__.py')) as f:
       match = REGEXP.search(f.read())
       if not match:
           raise InstallError("The Nipype __version__ variable was not found")
       return match.group(1)


def requires():
    with open('requirements.txt') as f:
        return f.read().splitlines()


def dependency_links():
    """
    Returns the non-PyPI ``qipipe`` requirements in
    ``constraints.txt`` which match the :const:`VCS_RQMT_PAT`
    pattern.

    :return: the non-PyPI package specifications
    """
    # FIXME - broken: install from PyPi fails on dcmstack.
    # Work-around is to clone qipipe repo and install with
    # --constraint constraints.txt option, as described in
    # doc/index.rst.
    #
    # TODO - revisit this yet again and eliminate doc step
    # if fixed.
    with open('constraints.txt') as f:
        rqmts = f.read().splitlines()
        # Match on git dependency links. Not a general solution, but
        # good enough so far.
        # TODO - revisit if Python 3 settles on a sane package manager.
        return [rqmt for rqmt in rqmts if VCS_RQMT_PAT.match(rqmt)]


def readme():
    with open("README.rst") as f:
        return f.read()


setup(
    name = 'qipipe',
    version = version('qipipe'),
    author = 'OHSU Knight Cancer Institute',
    author_email = 'loneyf@ohsu.edu',
    platforms = 'Any',
    license = 'MIT',
    keywords = 'Imaging QIN OHSU DCE MR XNAT Nipype',
    packages = find_packages(exclude=['test**']),
    package_data = dict(qipipe=['conf/*.cfg']),
    scripts = glob.glob('bin/*'),
    url = 'http://qipipe.readthedocs.org/en/latest/',
    description = 'Quantitative Imaging Profile pipeline',
    long_description = readme(),
    classifiers = [
        'Development Status :: 5 - Production/Stable',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'Environment :: Console',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
    ],
    install_requires = requires(),
    dependency_links = dependency_links()
)
